----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    16:55:30 04/14/2016 
-- Design Name: 
-- Module Name:    SPI_interface - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

use IEEE.std_logic_arith.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;


-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity SPI_interface is
	 Generic ( dataLength : integer := 16;   -- Lnge des Datenpaketes
				  T_clk : integer := 10);		-- ns - Entspricht der Periodendauer von clk.
    Port ( clk : in  STD_LOGIC;		-- Takt 
           enable : in  STD_LOGIC;	-- SPI_interface aktivieren
			  writeData : in std_logic_vector(dataLength-1 downto 0); 	-- Daten, welche seriell ber SPI bertragen werden sollen. 
																							-- 1 Bit read/write; 3 Bits Adresse; 12 Bits Daten. 
			  readData : out std_logic_vector(dataLength-5 downto 0); 
			  rst : in  STD_LOGIC;		-- Reset
			  
           SCS : out  STD_LOGIC;		-- Chip-Select zur Auswahl des Driver ICs 
           SCLK : out  STD_LOGIC;	-- Takt zum Schreiben und Lesen von Daten ber SPI
           DATAtoDRIVER : out  STD_LOGIC;	-- Input fr Slave (dort SDATI), also Output von Master. Hier Input fr Driver IC, Output von FPGA
			  
           DATAfromDRIVER : in  STD_LOGIC		-- Output von Slave (dort SDATO), also Input fr Master. Hier Output von Driver IC, Input fr FPGA
			);
           
end SPI_interface;

----------------------------------------------------------------------
-- SPI_interface stellt eine SPI Schnittstelle zu dem Controll-IC DRV8711 dar. 
-- Die unten definierten Zeitkonstanten, welche zur sicheren Datenbertragung
-- eingehalten werden mssen, sind durch die gewhlte Implementierung automatisch
-- eingehalten. 
-- Beschreibung der IOs: 
-- clk: Taktsignal. Die Periodendauer dieses Taktsignals muss mit T_clk bergeben werden. 
-- enable: Signal, um die Datenbertragung (Lesen oder Schreiben) zu starten. 
-- writeData: Daten, welche seriell ber SPI bertragen werden sollen (1 Bit read/write; 3 Bits Adresse; 12 Bits Daten). 
-- 			  Die bergabe der Daten erfolgt fr gewhnlich durch serialData. 
-- readData: Die seriell bertragenen Lese-Daten werden in diesem Vektor gespeichert. 
-- rst: Reset
-- SCS: Chip-Select Signal zur Auswahl des Controll-ICs. Die Datenbertragung ist nur mglich, wenn 
-- 	  dieses Signal auf 1 ist. 
-- SCLK: Taktsignal zur Datenbertragung. Dieses Taktsignal wird nur erzeugt, wenn auch Daten bertragen 
--			werden sollen (entspricht SCS auf high). Ansonsten ist es low. Bei einer steigenden Flanke dieses 
-- 		Signals werden Daten geschrieben, bei einer fallenden Flanke Daten gelesen. 
-- DATAtoDRIVER: Serielle Schnittstelle Input fr Slave (dort SDATI), also Output von Master. 
-- 				  Hier Input fr Driver IC, Output von FPGA
-- DATAfromDRIVER: Serielle Schnittstelle Output von Slave (dort SDATO), also Input fr Master. 
-- 					 Hier Output von Driver IC, Input fr FPGA
----------------------------------------------------------------------


architecture Behavioral of SPI_interface is

	type clkState is (stop, high, low); 
	
	constant T_CYC : integer := 250; 	-- ns - (1) Clock cycle time, min, in ns
--	constant T_CLKH : integer := 25; 	-- ns - (2) Clock high time, min, in ns
--	constant T_CLKL : integer := 25; 	-- ns - (3) Clock low time, min, in ns
--	constant T_SU_SDATI : integer := 5; -- ns - (4) Setup time, SDATI to SCLK, min, in ns
--	constant T_H_SDATI : integer := 1; 	-- ns - (5) Hold time, SDATI to SCLK, min, in ns
--	constant T_SU_SCS : integer := 5; 	-- ns - (6) Setup time, SCS to SCLK, min, in ns
--	constant T_H_SCS : integer := 1; 	-- ns - (7) Hold time, SCS to SCLK, min, in ns
--	constant T_L_SCS : integer := 100; 	-- ns - (8) Inactive time, SCS (between writes and reads), min, in ns
--	constant T_D_SDATO : integer := 10; -- ns - (9) Delay time, SCLK to SDATO (during read), max, in ns
	
	constant T_SLEEP : integer := 1000000; -- ns - Wake time (SLEEPn inactive to high-side gate drive enabled), max, in ns
	constant T_RESET : integer := 10000;	-- ns - Delay from power up or RESETn high until serial interface functional, max, in ns
		
	signal SCLK_cycles : integer range 0 to (dataLength + 2); 	-- Zhler, wieviele SCLK Taktzyklen vergangen sind. 
	signal SCLK_state : clkState; 
	signal start_clk : std_logic; 
	signal SCLK_timer : integer range 0 to T_CYC/2;
	
	signal SCS_int : std_logic; 
	signal clk_SetupWriteData : std_logic; 
	signal clk_SetupReadData : std_logic; 
	
--	signal SCLK_shift_reg : std_logic_vector(2 downto 0); 
	signal SCLK_shift_reg : std_logic_vector(4 downto 0); 
	signal writeData_reg : std_logic_vector(dataLength-1 downto 0); 
	signal dataBits : integer range 0 to dataLength; 
	
	
begin



----------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------
	-- Erzeugung des Taktes zur Datenbertragung SCLK
	-- 
	-- Der Takt selbst muss die Zeitanforderungen der Konstanten (1) - (3) erfllen. 
	-- Der Takt soll nur dann erzeugt werden, wenn auch Daten bertragen werden soll. Deshalb 
	-- startet der Takt erst, wenn das enable-Signal gesetzt wird. 
	-- Im nchsten Prozess wird das Chip-Select Signal SCS erzeugt. Dieses wird sofort
	-- mit enable gesetzt. Um die Zeitanforderung der Konstanten (6) zu erfllen, darf 
	-- die erste (steigende oder fallende) Taktflanke von SCLK nicht sofort mit dem 
	-- enable-Signal erfolgen. Es muss erst mindestens die Zeit T_SU_SCS abgewartet werden. 
	-- 
	-- Des Weiteren mssen beim Schreiben die Setup- und Hold-Zeiten (4) und (5) der Daten 
	-- im Hinblick auf den Takt eingehalten werden sowie beim Lesen die Delay time (9). Um 
	-- diese Zeiten zu bercksichtigen, werden insgesamt drei Taktsignale erzeugt: 
	-- clk_SetupWriteData, clk_SetupReadData, SCLK. Der Takt clk_SetupWriteData bestimmt, wann
	-- intern die Daten zum Schreiben angelegt werden. Der Takt clk_SetupReadData bestimmt, wann
	-- die Daten vom Slave gelesen werden knnen (DATAfromDRIVER). Der Takt SCLK entspricht dem
	-- eigentlichen Taktsignal, welches der Master erzeugt. Es kommt nach clk_SetupWriteData, 
	-- um die Konstante (4) einzuhalten und vor clk_SetupReadData, um die Konstante (9) einzuhalten. 
	-- Die Umsetzung erfolgt in zwei Prozessen. Im ersten Prozess wird clk_SetupWriteData erzeugt, 
	-- im zweiten Prozess werden ber ein Schieberegister die beiden Takte SCLK und clk_SetupReadData
	-- verzgert aus clk_SetupWriteData erzeugt. Der erste Prozess ist wie folgt aufgebaut: 
	-- 
	-- Fr die clk_SetupWriteData werden drei Zustnde eingefhrt: high, low, stop. Im Zustand high ist 
	-- clk_SetupWriteData = 1, in den anderen beiden ist clk_SetupWriteData = 0. Vom Zustand low geht es 
	-- nur wieder in den high Zustand. Vom high Zustand geht es in den low Zustand, wenn die Anzahl 
	-- der erzeugten Taktzyklen noch kleiner ist als die vorgegebene Datenlnge dataLength. 
	-- Ansonsten (Anzahl erzeugter Taktzyklen grer gleich dataLength) geht es in den stop Zustand. Dieser 
	-- wird bis zum nchsten enable Signal gehalten. 
	-- Der Zustandswechsel erfolgt durch Abfrage eines Zhlers SCLK_timer. Es wird davon ausgegangen, dass der
	-- interne Takt des FPGAs schneller ist, als der ber T_CYC definierte maximal erlaubte Takt SCLK. 
	-- Bei der unteren Umsetzung werden damit automatisch die erforderlichen Zeitkonstanten (1), (2) und (3) 
	-- eingehalten. Die Umsetzung geschieht wie folgt: 
	-- Nach einem enable Signal wird der Zhler SCLK_timer um die Periodendauer des internen Taktes T_clk (als
	-- generic bergeben) hochgezhlt. Wird die halbe Periodendauer des maximal erlaubten Taktes SCLK, also 
	-- T_CYC/2 berschritten, findet ein Pegelwechsel bei clk_SetupWriteData statt (von low auf high bzw. von high auf low). 
	-- Gleichzeitig wird gezhlt, wieviele Taktzyklen von clk_SetupWriteData bereits erzeugt wurden (SCLK_cycles). Wurden die 
	-- fr die vorgegebene Datenlnge erforderlichen Taktzyklen erzeugt, erfolgt ein letzter Pegelwechsel von high 
	-- auf low und der Zustand stop wird erreicht. 
	-- In einem weiteren Prozess wird das Chip-Select Signal SCS gesetzt. Um die Zeitkonstanten (6) und (7)
	-- ebenfalls einzuhalten, wird der Zhler SCLK_cycles im Zustand stop jedoch noch weiter hochgezhlt. Das 
	-- SCS-Signal wird mit enable gesetzt und sobald SCLK_cycles den Wert (dataLength + 1) erreicht wieder 
	-- zurckgesetzt. Damit sind die beide Zeitkonstanten (6) und (7) ebenfalls sicher eingehalten. 
	-- 
	-- Der Zhler dataBits gibt an, welches Bit gerade gelesen oder geschrieben werden kann/soll. Er wird in 
	-- dem Prozess zur bertragung der Daten verwendet. 
	-- 
	-- Die Zeitkonstante (8) sorgt fr einen Abstand zwischen Lese- und Schreibzyklen. Diese muss auerhalb 
	-- bercksichtigt werden. 
	
	process
		
	begin
		wait until rising_edge(clk);

		if enable = '1' then 	-- Prozess wird mit enable gestartet
			start_clk <= '1'; 	-- Signal zeigt an, dass ein enable-Signal gekommen ist und bleibt so lange gesetzt,
										-- bis die Daten bertragen/empfangen wurden. 
			SCLK_state <= low; 
			SCLK_timer <= 0; 
			SCLK_cycles <= 0; 
			dataBits <= dataLength;
			
		end if; 
		
		if start_clk = '1' then 
			
			CASE SCLK_state IS 	-- State-Machine 
				WHEN low => 		-- Zustand low --> SCLK geht/bleibt auf 0, bis die entsprechenden Zeiten erreicht sind. 
										-- Im Anschluss ist nur ein bergang in den Zustand high mglich. 
					clk_SetupWriteData <= '0'; 
					if SCLK_timer >= (T_CYC/2 - T_clk) then
						SCLK_timer <= 0; 
						SCLK_state <= high; 
						SCLK_cycles <= SCLK_cycles + 1; 
						dataBits <= dataBits - 1;
						
					else
						SCLK_timer <= SCLK_timer + T_clk;
					end if; 
				 
				WHEN high => 		-- Zustand high --> SCLK geht/bleibt auf 1, bis die entsprechenden Zeiten erreicht sind. 
										-- Im Anschluss geht es entweder in den Zustand low, oder, falls alle Daten bertragen/empfangen
										-- wurden, in den Zustand stop. 
					clk_SetupWriteData <= '1'; 
					if SCLK_timer >= (T_CYC/2 - T_clk) then
						SCLK_timer <= 0; 
						
						if SCLK_cycles = dataLength then 
							SCLK_state <= stop; 
						else 
							SCLK_state <= low; 
						end if; 
						
					else
						SCLK_timer <= SCLK_timer + T_clk;
					end if; 
					
				WHEN stop => 		-- Zustand stop --> SCLK bleibt auf 0, bis das nchste enable-Signal kommt. 
										-- Der timer wird trotzdem noch hochgezhlt, um fr den nchsten Prozess mit SCS 
										-- die Zeiten einzuhalten. 
					clk_SetupWriteData <= '0';
					
					if (SCLK_cycles >= dataLength) AND (SCLK_cycles <= dataLength + 1) then
						if SCLK_timer >= (T_CYC/2 - T_clk) then
							SCLK_timer <= 0; 
							SCLK_cycles <= SCLK_cycles + 1; 
							
						else
							SCLK_timer <= SCLK_timer + T_clk;
						end if; 
						
					else 
						start_clk <= '0';  
						SCLK_timer <= 0; 
						SCLK_cycles <= 0; 
						dataBits <= dataLength;
					end if; 
					
			END CASE; 
			
		end if; 
			
			
		if rst = '1' then 
			clk_SetupWriteData <= '0'; 
			start_clk <= '0'; 
			SCLK_state <= stop;
			SCLK_cycles <= 0;
			SCLK_timer <= 0; 
			dataBits <= 0; 

		end if; 
	end process; 
	
	
	process
	begin
	wait until rising_edge(clk); 

		if SCS_int = '1' then
			SCLK_shift_reg <= SCLK_shift_reg(SCLK_shift_reg'left-1 downto 0) & clk_SetupWriteData;
			SCLK <= SCLK_shift_reg(SCLK_shift_reg'left-3);

			if SCLK_shift_reg(SCLK_shift_reg'left) = '1' AND SCLK_shift_reg(SCLK_shift_reg'left-1) = '0' then
				clk_SetupReadData <= '1'; 
			else 
				clk_SetupReadData <= '0'; 
			end if; 
		end if; 
			
		if rst = '1' then 
			SCLK_shift_reg <= (others => '0'); 
			SCLK <= '0'; 
			clk_SetupReadData <= '0'; 
		end if; 
	end process; 
----------------------------------------------------------------------------------------



----------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------
	-- Erzeugung des Chip-Select Signals SCS 
	-- 
	-- Das SCS-Signal soll sofort mit dem enable-Signal gesetzt werden. Nach der Datenbertragung bleibt das SCS-Signal
	-- noch eine gewisse Zeit gesetzt (bis SCLK_cycles = (dataLength + 1)), damit die Zeitkonstante (7) sicher eingehalten
	-- wird. 
	
	process
	begin
		wait until rising_edge(clk);

		if SCLK_cycles = (dataLength + 1) then 	-- Ist das Datenpaket empfangen/gesendet, soll das SCS-Signal wieder auf 0 gesetzt werden. 
			SCS_int <= '0'; 
		end if;
		
		if enable = '1' then 	-- Das SCS-Signal soll sofort mit dem enable-Signal gesetzt werden. 
			SCS_int <= '1'; 			-- berschreibt das vorherige if..., falls direkt wieder ein enable kommt. 
		end if; 		
			
		if rst = '1' then 
			SCS_int <= '0'; 			
		end if; 
	end process; 
	
	SCS <= SCS_int;
----------------------------------------------------------------------------------------



----------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------
	-- bertragung der Daten (Lesen oder Schreiben). 
	-- 
	-- Dieser Teil ist in drei Prozesse aufgeteilt: Im ersten Prozess werden die angelegten
	-- Daten writeData in ein Register bernommen. Die Daten beinhalten:
	-- 1 Bit read/write; 3 Bits Adresse; 12 Bits Daten. 
	-- 
	-- Da die ersten 4 Bit auch bei einer Lese-Operation bertragen werden mssen, werden 
	-- writeData im ersten Prozess bei jedem enable-Signal bernommen. 
	-- 
	-- Im zweiten Prozess werden die Daten seriell an den Ausgang DATAtoDRIVER bergeben. 
	-- Dieser Prozess ist mit clk_SetupWriteData getaktet, um die Setup-Zeit einzuhalten. 
	-- Da sowohl bei einer Lese- als auch bei einer Schreib-Operation Daten vom Master
	-- gesendet werden mssen (immer read/write-Bit und Adresse), erfolgt hier keine 
	-- explizite Abfrage nach dem read/write-Bit, um welche Art von Operation es sich 
	-- handeln soll. Es werden immer alle Daten bertragen. Bei einer Lese-Operation 
	-- ignoriert der Slave jedoch einfach die nach der Adresse bertragenen Daten. 
	-- 
	-- Im dritten Prozess werden die Daten bei einer Leseoperation gelesen. Dieser Prozess
	-- ist mit clk_SetupReadData getaktet, um die Delay time einzuhalten. Des Weiteren 
	-- wird nur gelesen, wenn es sich auch um eine Lese-Operation handelt (read/write Bit = 1). 
	-- Gelesen werden auch nur die Daten, d.h. die Bits, welche bei den Takten 11 downto 0 anliegen. 


	-- Daten, welche geschrieben werden sollen, in ein Register bernehmen. 
	
	process
	begin
		wait until rising_edge(clk);
		
		if enable = '1' then 	
			writeData_reg <= writeData; 
		end if; 
		
		if rst = '1' then 
			writeData_reg <= (others => '0'); 		
		end if; 
	end process; 
	
	
	
-- Schreiben von Daten mit clk_SetupWriteData

	process
	begin
		wait until rising_edge(clk);

		if clk_SetupWriteData = '1' then
			-- Bei jeder steigenden Flanke soll geschrieben werden! 
			-- Auch bei einer Lese-Operation mssen zunchst das read/write-Bit
			-- sowie die Lese-Adresse geschrieben werden. Da die Daten-Bits bei 
			-- einer Lese-Operation  einfach ignoriert werden, wird hier nicht weiter
			-- unterschieden und es werden immer alle Bits auf DATAtoDRIVER geschrieben. 
			DATAtoDRIVER <= writeData_reg(dataBits);
		end if; 
		
		if rst = '1' then 
			DATAtoDRIVER <= '0'; 	
		end if; 
	end process; 
	
	
	
-- Lesen von Daten mit clk_SetupReadData

	process
	begin
		wait until rising_edge(clk);

		if clk_SetupReadData = '1' then 
			if writeData_reg(writeData_reg'left) = '1' then -- READ --> Fallende Flanke von SCLK + read/write Bit = 1 
				if dataBits <= dataLength-5 then
					readData(dataBits) <= DATAfromDRIVER;
				end if; 
			end if; 
		end if; 
		
		if rst = '1' then 
			readData <= (others => '0'); 		
		end if;
	end process; 
----------------------------------------------------------------------------------------

	
end Behavioral;








