----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    16:36:41 08/30/2016 
-- Design Name: 
-- Module Name:    SPI_writeReg - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity SPI_writeReg is
	Generic ( dataLength : integer := 16;   -- Lnge des Datenpaketes.
				 T_clk : integer := 10);		-- ns - Entspricht der Periodendauer von clk.
	Port (
			clk : in  STD_LOGIC;		-- Takt.
			enable : in STD_LOGIC; 	-- Enable.
			rst : in STD_LOGIC;	-- Reset.
			selectController : in integer range 1 to 3;
			writeAllRegToDriver : in std_logic; 
			writeOneRegToDriver : in std_logic; 
			regAddress : in  std_logic_vector(2 downto 0);	-- Adresse zum Schreiben.
			dataToIntReg : in std_logic_vector(11 downto 0);	-- Daten, die in das Register regAddress geschrieben werden sollen. 
			
			SCS : out  STD_LOGIC;	-- Chip-Select Signal zur Auswahl des Controll-ICs. 
			SCLK : out  STD_LOGIC;	-- Taktsignal zur Datenbertragung ber SPI.
			DATAtoDRIVER : out  STD_LOGIC;	-- Serielle Schnittstelle Input fr Slave (dort SDATI), also Output von Master. 
														-- Hier Input fr Driver IC, Output von FPGA.
			writeDone : out std_logic 	-- Lese- oder Schreibzugriff fr das entsprechende Register abgeschlossen. 
			);
end SPI_writeReg;


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- Allgemeine Beschreibung der Funktion: 
-- 
-- Mit SPI_writeReg knnen entweder alle Register mit den im FPGA hinterlegten Werten
-- oder ein einzelnes Register mit der Adresse regAddress und den Daten dataToIntReg
-- geschrieben werden.  
-- Im Moment ist hier kein Lesen der Register umgesetzt, da dies derzeit nicht 
-- bentigt wird. Es ist nur mglich die Register zu schreiben! 
-- 
-- Genauere Beschreibung: 
-- Mit selectController wird einer der drei mglichen Treiber-ICs ausgewhlt. 
-- Ist bei einem enable-Signal das Signal writeAllRegToDriver ebenfalls auf '1', dann 
-- werden alle Register des Treiber-ICs nacheinander mit den im FPGA hinterlegten 
-- Werten beschrieben. 
-- Ist bei einem enable-Signal das Signal writeOneRegToDriver auf '1' (und das Signal 
-- writeAllRegToDriver auf '0'), dann wird das Register des Treiber-ICs mit der 
-- Adresse regAddress mit den Daten dataToIntReg berschrieben. Gleichzeitig wird
-- das interne Register im FPGA mit der Adresse regAddress ebenfalls mit den Daten 
-- dataToIntReg berschrieben. 
-- Der Takt und die Daten fr die SPI-Schnittstelle werden ber SCLK und DATAtoDRIVER
-- zur Verfgung gestellt. Zustzlich wird das Chip-Select Signal ber SCS gesetzt. 
-- Mit dem Signal writeDone wird angezeigt, dass der Schreibprozess abgeschlossen ist. 
-- Werden alle Register geschrieben, wird writeDone nur nach dem letzten Register 
-- gesetzt. Wird nur ein Register geschrieben, wird es entsprechend nach dem 
-- Schreibprozess des einen Registers gesetzt.  
-- 
-- Beschreibung der Ein- und Ausgangssignale: 
-- clk : Taktsignal.
-- enable : Enable-Signal. Startet den Prozess. 
-- rst :  Reset.
-- selectController : Auswahl des Controllers zum Schreiben / Lesen. Es sind derzeit 3 Kontroller hinterlegt. 
-- writeAllRegToDriver : Ist dieses Signal bei einem enable-Signal auf '1', werden alle Register des Treiber-ICs 
-- 							 mit den im FPGA hinterlegten Werten geschrieben. 
-- writeOneRegToDriver : Ist dieses Signal bei einem enable-Signal auf '1' (und gleichzeitig writeAllRegToDriver auf '0')
-- 							 wird (nur)das Register mit der Adresse regAddress mit den Daten dataToIntReg beschrieben. Hierbei 
-- 							 werden sowohl das entsprechende Register im FPGA, als auch das entsprechende Register im Treiber-IC 
-- 							 auf die Werte von dataToIntReg gesetzt. 
-- regAddress : Adresse zum Schreiben, wenn nur ein ausgewhltes Register beschrieben werden soll. 
-- dataToIntReg : Daten, die in das Register regAddress geschrieben werden sollen, falls nur ein 
--						ausgewhltes Register beschrieben werden soll. 
-- SCS : Chip-Select Signal zur Auswahl des Treiber-ICs. 
-- SCLK : Taktsignal zur Datenbertragung ber SPI.
-- DATAtoDRIVER : Serielle Schnittstelle Input fr Slave (dort SDATI), also Output von Master. Hier Input fr Driver IC, Output von FPGA.
-- 					Die Schnittstelle dient zum schreiben von Daten auf den Treiber-IC. 
-- writeDone : Lese- oder Schreibzugriff fr das entsprechende bzw. alle Register abgeschlossen. 
------------------------------------------------------------------------------------
------------------------------------------------------------------------------------


architecture Behavioral of SPI_writeReg is
	
	constant T_L_SCS : integer := 100; 	-- ns - (8) Inactive time, SCS (between writes and reads), min, in ns
	constant T_waitAfterWrite : integer := (T_L_SCS/T_clk) + 1;
	
	
	signal SPI_readWriteReg_enable : std_logic; 
	signal regAddress_int : std_logic_vector(2 downto 0);
	signal readOrWriteIntReg : std_logic; 
	signal readWriteDone : std_logic; 
	
	signal selectController_int : integer range 1 to 3 := 1;
	signal dataToIntReg_int : std_logic_vector(11 downto 0);
	
	signal writeAll : std_logic; 
	signal SPI_waitBetweenWriteCycles : std_logic_vector(T_waitAfterWrite downto 0); 	-- Lnge des Schieberegisters richtet sich nach der 
																													-- Zeitkonstanten (8) zur Einhaltung einer Pause zwischen
																													-- Lese- und Schreibzyklen. 

	component SPI_readWriteReg is
	Generic ( dataLength : integer;   -- Lnge des Datenpaketes.
				 T_clk : integer);		-- ns - Entspricht der Periodendauer von clk.
	port(
			clk : in  STD_LOGIC;		-- Takt.
			enable : in STD_LOGIC; 	-- Enable.
			rst : in STD_LOGIC;	-- Reset.
			selectController : in integer range 1 to 3;	-- Auswahl des Controllers zum Schreiben / Lesen. 
			readWriteBit_DriverIC : in STD_LOGIC; 	-- Lese- oder Schreibzugriff auf den Motor-Treiber IC
			readOrWriteIntReg : in STD_LOGIC; 	-- Lese- oder Schreibzugriff auf die im FPGA internen Datenregiser fr den Treiber-IC
															-- readOrWriteIntReg = '1' --> Set = Schreibe dataIn in das Register von selectController an die Stelle von address
															-- readOrWriteIntReg = '0' --> Read = Nimm die Daten aus dem Register von selectController an der Stelle von address
															--										        und gib sie an dataOut weiter. 
			regAddress : in  std_logic_vector(2 downto 0);	-- Adresse zum Schreiben (/Lesen).
			dataToIntReg : in std_logic_vector(11 downto 0);	-- Daten, die in das Register regAddress geschrieben werden sollen. 
			DATAfromDRIVER : in  STD_LOGIC;	-- Serielle Schnittstelle Output von Slave (dort SDATO), also Input fr Master. 
														-- Hier Output von Driver IC, Input fr FPGA.
			SCS : out  STD_LOGIC;	-- Chip-Select Signal zur Auswahl des Controll-ICs. 
			SCLK : out  STD_LOGIC;	-- Taktsignal zur Datenbertragung ber SPI.
			DATAtoDRIVER : out  STD_LOGIC;	-- Serielle Schnittstelle Input fr Slave (dort SDATI), also Output von Master. 
														-- Hier Input fr Driver IC, Output von FPGA.
			SPI_readData : out std_logic_vector(dataLength-5 downto 0); 	-- Gelesene Daten
			readWriteDone : out std_logic 	-- Lese- oder Schreibzugriff fr das entsprechende Register abgeschlossen. 
			); 
	end component;



begin

	
	SPI_readWriteReg_inst : SPI_readWriteReg
	Generic map ( 	dataLength => dataLength, 
						T_clk => T_clk)
	port map (
			clk => clk,
			enable => SPI_readWriteReg_enable,
			rst => rst,	
			selectController => selectController_int,
			readWriteBit_DriverIC => '0',		-- 0 fr Write; 1 fr Read.
			readOrWriteIntReg => readOrWriteIntReg,
			regAddress => regAddress_int,
			dataToIntReg => dataToIntReg_int,
			DATAfromDRIVER => '0',
			SCS => SCS,
			SCLK => SCLK,
			DATAtoDRIVER => DATAtoDRIVER,
			SPI_readData => open,
			readWriteDone => readWriteDone
			); 
	
	
	
	process
	begin
		wait until rising_edge(clk);
		
		SPI_readWriteReg_enable <= '0'; 
		writeDone <= '0';

		if enable = '1' then 
		
			-- Wenn alle Register geschrieben werden sollen, wird das Signal writeAll auf '1' gesetzt. 
			-- Gleichzeitig wird die Adresse fr das erste Register bergeben und das enable-Signal
			-- gesetzt. 
			-- Im Fall, dass nur ein Register geschrieben werden soll, wird writeAll auf '0' gesetzt. 
			-- Gleichzeitg werden das enable-Signal gesetzt, die Adresse des zu schreibenden Registers 
			-- und dirket bei der Komponenten-Instantiierung von SPI_readWriteReg die Daten bernommen. 
			
			selectController_int <= selectController; 
			
			if writeAllRegToDriver = '1' then 
				
				regAddress_int <= "000";	-- starte mit Hex Adresse 00 (CTRL)
				readOrWriteIntReg <= '0';	-- Zeigt an, dass die im FPGA hinterlegten Werte ausgewhlt werden sollen. 
													-- Damit werden regAdress und dataToIntReg ignoriert. 
				SPI_readWriteReg_enable <= '1'; 
				
				writeAll <= '1'; 
				SPI_waitBetweenWriteCycles <= (others => '0'); 
				
			elsif writeOneRegToDriver = '1' then 
				
				regAddress_int <= regAddress; 
				dataToIntReg_int <= dataToIntReg; 
				
				readOrWriteIntReg <= '1';
				SPI_readWriteReg_enable <= '1'; 
				
				writeAll <= '0';
				SPI_waitBetweenWriteCycles <= (others => '0'); 
				
			end if; 
		end if; 
		
		-- Da sofort mit dem enable-Signal der Schreib-Prozess ausgefhrt wird, wird mit Abschluss dieses Schreib-Prozesses
		-- das Signal readWriteDone einmal gesetzt. In dem hier verwendeten Schieberegister wird dieser Wert bis zum linken
		-- Ende verschoben. Dies entspricht einer Verzgerung, die im Datenblatt des Treiber-ICs zwischen zwei Schreib-Zyklen
		-- eingehalten werden muss. Wurde diese Zeit abgewartet, wird die nachfolgende if-Abfrage ausgefhrt. In dieser wird 
		-- nochmals unterschieden, ob alle oder nur ein Register geschrieben werden sollen. Sollen alle Register geschrieben
		-- werden, wird das nchste Register ausgewhlt und ein neuer Schreib-Zyklus gestartet. Nach dem letzten Register 
		-- wird mit writeDone angezeigt, dass alle Register geschrieben wurden. Sollte nur ein Register geschrieben werden, 
		-- wird direkt writeDone gesetzt und damit angezeigt, dass der Schreib-Prozess abgeschlossen ist. 
		SPI_waitBetweenWriteCycles <= SPI_waitBetweenWriteCycles(SPI_waitBetweenWriteCycles'left-1 downto 0) & readWriteDone; 
		
		if SPI_waitBetweenWriteCycles(SPI_waitBetweenWriteCycles'left downto SPI_waitBetweenWriteCycles'left-1) = "10" then 
			if writeAll = '1' then 					
			
				SPI_waitBetweenWriteCycles <= (others => '0'); 
			
				CASE regAddress_int IS 
					WHEN "000" => 
						regAddress_int <= "001";
						SPI_readWriteReg_enable <= '1'; 
						
					WHEN "001" => 
						regAddress_int <= "010";
						SPI_readWriteReg_enable <= '1'; 
						
					WHEN "010" => 
						regAddress_int <= "011";
						SPI_readWriteReg_enable <= '1'; 
						
					WHEN "011" => 
						regAddress_int <= "100";
						SPI_readWriteReg_enable <= '1'; 
						
					WHEN "100" => 
						regAddress_int <= "101";
						SPI_readWriteReg_enable <= '1'; 
						
					WHEN "101" => 
						regAddress_int <= "110";
						SPI_readWriteReg_enable <= '1'; 
						
					WHEN "110" => 
						regAddress_int <= "111";
						SPI_readWriteReg_enable <= '1'; 
						
					WHEN "111" => 
						writeDone <= '1'; 
						
					WHEN others => 
						null; 
						
				END CASE; 
				
			else 	-- writeAll = '0'
			
				writeDone <= '1'; 
			
			end if; 
		end if; 
			
			
		if rst = '1' then 
			writeAll <= '0';
			SPI_readWriteReg_enable <= '0'; 
			SPI_waitBetweenWriteCycles <= (others => '0'); 
			
			writeDone <= '0'; 
		end if; 
	end process; 
	
	

	
end Behavioral;




